import urllib.request
import pickle

def request(url , ip = "127.0.0.1" , port = "8000"):
	full_url = "http://{ip}:{port}/{url}".format(ip = ip , port = port , url = url)
	response = urllib.request.urlopen(full_url).read()
	return pickle.loads(response)

class Requester:
	def __init__(self , ip = "127.0.0.1" , port = "30726"):
		self.ip = ip
		self.port = port
	def request(self , url):
		return request(url , self.ip , self.port)


class ArgumentError(Exception):
	'''函数的参数不符合要求'''
	def __init__(self, func_name , var_name = None , var_val = None , class_name = None , note_str = None):
		''' 
		参数：
			func_name： 字符串。抛出异常的位置
			var_name： str，抛出异常的变量名
			var_val： object，抛出异常的变量值
			in_class： 如果不是None，就是某个class
		'''

		if class_name is not None:
			func_name = "{class_name}.{func_name}".format(class_name = class_name , func_name = func_name)

		if note_str is None:
			note_str = "Invalid argument value"
		
		var_str = ""
		if var_name is not None:
			var_str = ": {var_name} can not be {var_val}".format(var_name = var_name , var_val = var_val)
		
		self.args = ["{func_name}: {note_str}{var_str}".format(
			func_name = func_name , note_str = note_str , var_str = var_str
		)]
		
class YAttributeError(Exception):
	'''函数的参数不符合要求'''
	def __init__(self, func_name , var_name = None , var_val = None , class_name = None , note_str = None):
		''' 
		参数：
			func_name： 字符串。抛出异常的位置
			var_name： str，抛出异常的变量名
			var_val： object，抛出异常的变量值
			in_class： 如果不是None，就是某个class
		'''

		if class_name is not None:
			func_name = "{class_name}.{func_name}".format(class_name = class_name , func_name = func_name)

		if note_str is None:
			note_str = "Invalid artribute value"
		
		var_str = ""
		if var_name is not None:
			var_str = ": {var_name} can not be {var_val}".format(var_name = var_name , var_val = var_val)
		
		self.args = ["{func_name}: {note_str}{var_str}".format(
			func_name = func_name , note_str = note_str , var_str = var_str
		)]



import torch as tc

class EnsembleModel:
	'''Ensemble模型

	只可用于测试，不要训练这个模型。
	另外注意这个类会自动取softmax，所以要保证模型中没有取softmax，loss函数中也不会取softmax

	参数：
	models：要混合的模型
	device：在哪个设备上混合。
		其实对小模型无所谓的，有这个选项主要是允许把model都放到cpu上，每次只取一个model在gpu上生成，以节省
		显存。
	'''

	def __init__(self , models , device = 0):
		self.models = models
		self.device = device

	def forward(self , *pargs , output_preds = False , **kwargs):
		models = self.models
		device = self.device

		with tc.no_grad():
			preds = [0 for _ in range(len(models))]
			for i , model in enumerate(models):

				old_device = next(model.parameters()).device
				model = model.to(device)
				preds[i] = model(*pargs , **kwargs)
				model = model.to(old_device) #如果他本来在cpu上，生成完之后还是把他放回cpu

		if output_preds:
			return preds

		pred = 0
		for x in preds:
			pred = pred + tc.softmax(x , dim = -1)
		pred /= len(models)

		return pred

	def __call__(self , *pargs , **kwargs):
		return self.forward(*pargs , **kwargs)

	@property
	def parameters(self): #for device deciding
		return self.models[0].parameters

	def to(self , device):
		self.device = device
		return self

	def eval(self):
		for i in range(len(self.models)):
			self.models[i] = self.models[i].eval()
		return self
	def train(self):
		for i in range(len(self.models)):
			self.models[i] = self.models[i].train()
		return self




from .fakepath import new_fakefolder
from ..universe.onexit import add_quit_methods
import sqlite3
import os
from collections import Iterable
import pickle
from ..universe.exceptions import ArgumentError , YAttributeError

HASH_FOLDER_NAME = "YToolsHash"

class StaticHash:

	def __init__(self , db_name , tb_name = "Hash" ,  keysize = 255):

		self.tb_name = tb_name

		# 获取数据库地址
		folderpath = new_fakefolder( HASH_FOLDER_NAME )
		self.db_path = os.path.join( folderpath , db_name )

		# 连接数据库
		self.connection = sqlite3.connect(self.db_path , check_same_thread = False)
		self.cursor = self.connection.cursor()
		self.closed = False
		add_quit_methods(self.close) #确保退出时关闭

		# 初始化数据库
		self.keysize = keysize
		self.init()

	def close(self):
		if self.closed: #防止反复关闭
			return
		self.closed = True

		self.cursor.close()
		self.connection.close()

	def delete(self):
		'''!! 删除所有记录'''
		self.close()
		os.remove(self.db_path)

	def init(self):
		'''初始化表'''
		self.cursor.execute("""
			CREATE TABLE IF NOT EXISTS {tablename}
			(
				key varchar({keysize}) PRIMARY KEY, 
				val int
			);
		""".format(tablename = self.tb_name, keysize = self.keysize))
		self.connection.commit()

	def get(self , key):
		'''查询某个key的值，没有则返回None'''

		ret = self.cursor.execute("""
			SELECT val
			FROM {tablename}
			WHERE key = '{key}';
		""".format(tablename = self.tb_name , key = key)).fetchall()

		if len(ret) == 0:
			return None

		return ret[0][0]

	def set(self , key , val , commit = True):
		'''将某个key设为val'''

		#insert或replace
		self.cursor.execute("""
			INSERT OR REPLACE INTO {tablename}
			VALUES ('{key}',{val});
		""".format(tablename = self.tb_name , key = key , val = val))

		if commit:
			self.connection.commit()

	def ensure(self , key , val , commit = True):
		'''如果某个key不存在，则将其设为val'''

		self.cursor.execute("""
			INSERT OR IGNORE INTO {tablename}
			VALUES ('{key}',{val}); 
		""".format(tablename = self.tb_name , key = key , val = val))

		if commit:
			self.connection.commit()

	def plus(self , key , val , commit = True):
		'''将key对应的值加val，并返回加之后的值。需要用户确保其存在'''

		#先update
		self.cursor.execute("""
			UPDATE {tablename}
			SET val = val + {val}
			WHERE key = '{key}';
		""".format(tablename = self.tb_name , key = key , val = val))

		ret_val = self.get(key) #在commit前查询，防止被其他线程修改

		if commit:
			self.connection.commit()

		return ret_val

	def commit(self):
		self.connection.commit()


class DoubleHash(StaticHash):
	'''双向哈希，key和val都是unique的'''

	def __init__(self , db_name , tb_name = "DoubleHash" ,  keysize = 255):
		super().__init__(db_name , tb_name , keysize)

	def init(self):
		'''初始化表'''
		self.cursor.execute("""
			CREATE TABLE IF NOT EXISTS {tablename}
			(
				key varchar({keysize}) PRIMARY KEY, 
				val int NOT NULL UNIQUE 
			);
		""".format(tablename = self.tb_name, keysize = self.keysize))
		self.connection.commit()

	def ask_val(self , key):
		return self.get(key)

	def ask_key(self , val):
		'''查询某个val对应的key的的值，没有则返回None'''
		ret = self.cursor.execute("""
			SELECT key
			FROM {tablename}
			WHERE val = {val};
		""".format(tablename = self.tb_name , val = val)).fetchall()

		if len(ret) == 0:
			return None

		return ret[0][0]

	def ask(self , key = None, val = None):
		if key is not None:
			return self.ask_val(key)
		if val is None:
			return None
		return self.ask_key(val)


class HighDimHash(StaticHash):
	'''高维哈希，前m维是键，后n维是值'''

	def __init__(self , db_name , tb_name = "HighDimHash", keydim = 2 , valdim = 1 , always_tuple = False):
		'''
			keydim：键空间的维度
			valdim：值空间的维度
			always_tuple：如果为False，则在只有值只有一个元素时返回值而不是元组。默认False
		'''


		self.always_tuple = always_tuple

		self.keydim = keydim
		self.valdim = valdim

		self.key_list = ["key_{idx}".format(idx = i) for i in range(self.keydim)]
		self.val_list = ["val_{idx}".format(idx = i) for i in range(self.valdim)]

		super().__init__(db_name , tb_name , keysize = None)

	def name_value_list(self , namelist , vallist , sep = ","):
		'''将一系列名和一系列值转化为像『a=xx,b=xx』这样的字符串'''
		return sep.join( ["%s=%d" % (nam,val) for nam,val in zip(namelist , vallist)] )

	def init(self):
		'''初始化表'''
		self.cursor.execute('''
			CREATE TABLE IF NOT EXISTS {tablename}
			(
				{keys},
				{vals},
				PRIMARY KEY({primary_keys})
			);
		'''.format(tablename = self.tb_name, 
			keys = ",\n".join( ["%s int" % x for x in self.key_list] ) , 
			vals = ",\n".join( ["%s int" % x for x in self.val_list] ) , 
			primary_keys = ",".join(self.key_list), 
		))
		self.connection.commit()

	def get(self , keys):
		'''
			注意此处keys应是m维列表
			查询某个keys的值，没有则返回None
		'''

		keys , _ = self.check_kv(keys)

		if len(keys) != self.keydim:
			raise ArgumentError("get" , "len(keys)" , len(keys) , 
				class_name = "HighDimHash" , note_str = "key dim invalid")

		ret = self.cursor.execute("""
			SELECT {vals}
			FROM {tablename}
			WHERE {key_list};
		""".format(tablename = self.tb_name , 
			vals 	 = ",".join(self.val_list) ,
			key_list = self.name_value_list(self.key_list , keys , sep = " AND ") , 
		)).fetchall()

		if len(ret) == 0:
			return None

		if self.valdim == 1 and not self.always_tuple:
			return ret[0][0] # 返回唯一的一个值
		return ret[0] #返回元组

	def get_partial(self , keys):
		'''
			注意此处keys应是m维列表
			此处keys中的某些维度允许是None，表示查询这一维度任意的结果。
			返回一个列表表示所有的结果，列表元素是(keys , vals)
		'''

		keys , _ = self.check_kv(keys)

		if len(keys) != self.keydim:
			raise ArgumentError("get_partial" , "len(keys)" , len(keys) , 
				class_name = "HighDimHash" , note_str = "key dim invalid")

		ret = self.cursor.execute("""
			SELECT {keys} , {vals}
			FROM {tablename}
			WHERE {key_list};
		""".format(tablename = self.tb_name , 
			vals = ",".join(self.val_list) ,
			keys = ",".join(self.key_list) ,
			key_list = " AND ".join( ["%s=%d" % (nam,val) for nam,val in zip(self.key_list , keys) if val is not None] )
		)).fetchall()

		ret = [ [ x[:self.keydim] , x[self.keydim:] ] for x in ret] # 把keys分为一组，vals分为一组

		return ret #返回元组

	def check_kv(self , keys , vals = None):
		if len(keys) != self.keydim:
			raise RuntimeError("key dim invalid")


		if vals is not None:
			if self.valdim == 1 and not isinstance(vals , Iterable):
				vals = [vals] 

			if len(vals) != self.valdim:
				raise ArgumentError("check_kv" , "len(keys)" , len(keys) , 
					class_name = "HighDimHash" , note_str = "key dim invalid")

		return keys , vals

	def set(self , keys , vals , commit = True):
		'''将某个key设为val'''
		keys , vals = self.check_kv(keys , vals)

		self.cursor.execute("""
			INSERT OR REPLACE INTO {tablename}
			VALUES ({keys},{vals}); 
		""".format(tablename = self.tb_name , 
			keys = ",".join([str(x) for x in keys]) , 
			vals = ",".join([str(x) for x in vals]) , 
		))

		if commit:
			self.connection.commit()

	def ensure(self , keys , vals , commit = True):
		'''如果某个key不存在，则将其设为val'''
		keys , vals = self.check_kv(keys , vals)

		self.cursor.execute("""
			INSERT OR IGNORE INTO {tablename}
			VALUES ({keys},{vals}); 
		""".format(tablename = self.tb_name , 			
			keys = ",".join([str(x) for x in keys]) , 
			vals = ",".join([str(x) for x in vals]) , 
		))

		if commit:
			self.connection.commit()

	def plus(self , keys , vals , commit = True):
		'''将key对应的值加val，并返回加之后的值。需要用户确保其存在'''
		
		if self.valdim != 1:
			raise YAttributeError("plus" , "valdim" , self.valdim , 
				class_name = "HighDimHash" , note_str = "Highdim Table can only plus when valdim == 1")

			raise RuntimeError("HIghdim Table can only plus when valdim == 1")

		keys , vals = self.check_kv(keys , vals)
		val = vals[0]

		#先update
		self.cursor.execute("""
			UPDATE {tablename}
			SET {val_name} = {val_name} + {val}
			WHERE {key_list};
		""".format(tablename = self.tb_name , val_name = self.val_list[0] , val = val , 
			key_list = self.name_value_list(self.key_list , keys , sep = " AND ") 
		))

		ret_val = self.get(keys) #在commit前查询，防止被其他线程修改

		if commit:
			self.connection.commit()

		return ret_val


class StaticBlob(StaticHash):

	def __init__(self , db_name , tb_name = "BlobHash" , keysize = 255):
		super().__init__(db_name , tb_name , keysize)

	def close(self):
		if self.closed: #防止反复关闭
			return
		self.closed = True

		self.cursor.close()
		self.connection.close()

	def init(self):
		'''初始化表'''
		self.cursor.execute("""
			CREATE TABLE IF NOT EXISTS {tablename}
			(
				key varchar({keysize}) PRIMARY KEY, 
				val blob
			);
		""".format(tablename = self.tb_name, keysize = self.keysize))
		self.connection.commit()

	def get(self , key):
		'''查询某个key的值，没有则返回None'''

		ret = self.cursor.execute("""
			SELECT val
			FROM {tablename}
			WHERE key = '{key}';
		""".format(tablename = self.tb_name , key = key)).fetchall()

		if len(ret) == 0:
			return None

		return pickle.loads(ret[0][0])

	def set(self , key , val , commit = True):
		'''将某个key设为val'''

		#insert或replace
		self.cursor.execute("""
			INSERT OR REPLACE INTO {tablename}
			VALUES ('{key}',?);
		""".format(tablename = self.tb_name , key = key) , [sqlite3.Binary(pickle.dumps(val))])

		if commit:
			self.connection.commit()

	def ensure(self , key , val , commit = True):
		'''如果某个key不存在，则将其设为val'''

		self.cursor.execute("""
			INSERT OR IGNORE INTO {tablename}
			VALUES ('{key}',?); 
		""".format(tablename = self.tb_name , key = key) , [sqlite3.Binary(pickle.dumps(val))])

		if commit:
			self.connection.commit()


	def commit(self):
		self.connection.commit()



import socket
import copy
import pdb
import threading
import random
from ..universe.onexit import add_quit_methods

MSG_MAX_LENGTH = 2048

def randport():
	return random.randint(23333 , 60000)

class SendServer:
	def __init__(self , host = "127.0.0.1"):
		self.host = host
		self.targets = {}

	def __del__(self):
		self.close()

	def add_target(self , ip = "127.0.0.1" , port = 65432):
		
		if self.targets.get( (ip , port) , None) is not None: #已经建立连接了
			return True

		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		try:
			s.connect( (ip, port) )		
		except ConnectionRefusedError:
			return False

		self.targets[ip , port] = s

		return True

	def remove_target(self , ip , port):
		self.targets[ip , port].close()
		self.targets.pop((ip , port))

	def send(self , data):
		'''返回所有无法响应的目标'''

		leaved = []
		for ip , port in copy.copy(self.targets):
			if not self.send_to(ip , port , data):
				leaved.append((ip , port))
		return leaved

	def send_to(self , ip , port , data):
		'''return false if connection closed
		'''
		leaved = False

		try:
			self.targets[(ip , port)].sendall(data)
		except ConnectionResetError:
			leaved = True
			
		return not leaved

	def close(self):
		for addr , soc in self.targets.items():
			soc.close()


def default_callback(data , addr , who_get = None):
	ip , port = addr
	print (ip , ":" , port , " says ", data)

class ChildListener(threading.Thread):
	def __init__(self , conn , addr , callback , parent):
		threading.Thread.__init__(self)
		self.setDaemon(True)

		self.conn = conn
		self.addr = addr
		self.callback = callback
		self.closed = False
		self.parent = parent

		self.tarip = None
		self.tarport = None #listen port
		
	def __del__(self):
		self.close()

	def run(self):
		with self.conn:
			while not self.closed:
				try:
					data = self.conn.recv(MSG_MAX_LENGTH)
				except ConnectionResetError:
					if self.parent.unexpect_quit:
						self.parent.unexpect_quit(self.tarip , self.tarport)
					break

				if not data:
					continue
				self.callback(data , self.addr , who_get = self)
		self.close()

	def close(self):
		self.closed = True

class ListenServer(threading.Thread):
	def __init__(self , host  = "127.0.0.1" , port  = 65432 , callback = default_callback , block = False):
		threading.Thread.__init__(self)
		self.setDaemon(True)

		self.host = host
		self.port = port
		self.callback = callback

		self.childs = []
		self.closed = False
	
		self.block = block

		self.unexpect_quit = None

		add_quit_methods(self.close)

	def __del__(self):
		self.close()

	def run(self):
		with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
			s.bind((self.host, self.port))
			s.listen(50)

			while not self.closed:
				conn, addr = s.accept()

				new_child = ChildListener(conn , addr , self.callback , parent = self)
				new_child.start()
				self.childs.append(new_child)

	def close_one(self , tarip , tarport):
		the_child = None
		for x in self.childs:
			if x.tarip == tarip and x.tarport == tarport:
				the_child = x
				break
		if the_child is None:
			return

		the_child.close()
		self.childs.remove(the_child)

	def close(self):
		for x in self.childs:
			x.close()
		self.closed = True



'''只要保证写的位置不重合，所有进程可以同时写文件！'''

def get_filesize(file):
	file.seek(0 , 2) #移到文件末尾
	return file.tell()

def write_file(file , pos , data , flush = True):
	file.seek(pos , 0) #从头偏移
	file.write(data)
	if flush:
		file.flush()
	return file

def read_file(file , pos , size):
	'''从pos开始读len个字节'''
	file.seek(pos , 0) #从头偏移
	return file.read(size)


'''注意，这个模块可能死锁'''

from ..fakepath import fakepath_abs , new_fakefolder
import os
import random
import os.path as P

my_id = "{pid}_{seed}".format( pid = os.getpid() , seed = random.randint(0,233333)) #进程相关的id
name = "YFILELOCK/"
new_fakefolder(name)

def get_path(foldername):
	'''给定lock的文件夹名，创建对应文件夹'''
	pt = fakepath_abs(P.join(name , foldername))
	new_fakefolder(pt)
	return pt

def checkfile(path , my_filename):
	'''参数是绝对路径，检查这个路径下面有没有.lock结尾的文件，返回True表示没有。'''
	return len( [ x for x in os.listdir(path) if x.endswith(".lock") and x != my_filename] ) == 0


def acquire_lock(foldername , my_name = "0"):
	my_filename = "{my_id}_{my_name}.lock".format(my_id = my_id , my_name = my_name) #生成自己的锁文件的文件名
	path = get_path(foldername) #锁目录
	my_file = P.join(path , my_filename) #自己的锁文件的文件路径

	while True:	
		# 等待其他进程释放锁
		while not checkfile(path , my_filename): pass

		# 上自己的锁
		open(my_file , "w").close()

		if not checkfile(path , my_filename): # 如果发现还有不是自己的锁
			os.remove(my_file)  # 就删掉自己的锁
			continue 			# 并重新检查

		break # 加锁成功

def release_lock(foldername , my_name = "0"):
	my_filename = "{my_id}_{my_name}.lock".format(my_id = my_id , my_name = my_name) #生成自己的锁文件的文件名
	path = get_path(foldername) #锁目录
	my_file = P.join(path , my_filename) #自己的锁文件的文件名
	os.remove(my_file)  # 删掉自己的锁



class FileLock:
	'''锁定一个文件（目录）'''

	def __init__(self , foldername , threadname = None):
		if threadname is None:
			threadname = str( random.randint(0,233333) ) #生成一个线程相关的随机数

		self.foldername = foldername
		self.threadname = threadname

	def __enter__(self):
		acquire_lock(self.foldername , self.threadname)
		return self

	def __exit__(self , *args , **kwargs):
		release_lock(self.foldername , self.threadname)





from django.urls import path
from django.http import HttpResponse
import pickle
import json

urlpatterns = []

def make_response_func(func , configs):
	def _resp_func(*pargs , **kwargs):
		resp = func(*pargs , **kwargs)

		if configs["encode"] == "pickle":
			resp = pickle.dumps(resp) #用pickle编码
		if configs["encode"] == "json":
			resp = json.dumps(resp) #用pickle编码
		
		resp = HttpResponse(resp)

		if configs["cross_domain"]:
			resp["Access-Control-Allow-Origin"]  = "*"
			resp["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
			resp["Access-Control-Max-Age"] 		 = "1000"
			resp["Access-Control-Allow-Headers"] = "*"

		return resp
	return _resp_func

def add_response(url , func , configs):
	urlpatterns.append(path(url , make_response_func(func , configs)))



'''
	让同一个cmd执行多条命令（代替os.system())
	沙雕版
'''
import os
from pymouse import PyMouse
from pykeyboard import PyKeyboard
import time

_mouse = PyMouse()
_key = PyKeyboard()

class Cmd:
	'''
		创建一个控制台对象，稍后可以对其提供指令
	'''
	def __init__(self , start_cmd_cm = ["start cmd"] , sleep_time = 0.3):
		'''
			创建一个控制台对象
			一条「指令」指的是一系列按键命令，由PyKeyboard来运行

			start_cmd_cm : 用于打开控制台的指令列表
			sleep_time : 每次执行指令的间隔时间（单位：秒）
				（因为指令实际上会并行执行，为了保证顺序执行，两个指令间要有一定的间隔）
		'''
		self.k = _key
		self.m = _mouse
		self.sleep_time = sleep_time
		self.start_cmd_cm = start_cmd_cm
		self.cm_lis = []

	def start(self):
		'''
			打开控制台，用初始化的时候提供的打开控制台指令
		'''
		for cm in self.start_cmd_cm:
			os.system(cm)
			time.sleep(0.1)

	def decide_func(self , cm):
		'''
			抉择对于一个指令应该用哪一个函数来运行之
		'''
		if isinstance(cm , str):
			if len(cm) > 1:
				return self.k.type_string
			return self.k.press_key
		elif isinstance(cm , list):
			return self.k.press_keys
		return self.k.press_key

	def add_cm(self , cm , sleep_time = -1):
		'''
			向指令列表中加入一条指令

			sleep_time : 运行完本次指令后的停留时间，如果给定的值<0或者留空则用默认值
		'''
		self.cm_lis.append( (cm , sleep_time) )

	def run_cm(self , cm , sleep_time = -1):
		'''
			立刻执行一条指令

			sleep_time : 运行完本次指令后的停留时间，如果给定的值<0或者留空则用默认值
		'''
		func = self.decide_func(cm)
		func(cm)

		if sleep_time < 0:
			time.sleep(self.sleep_time)
		else:
			time.sleep(sleep_time)

	def run(self , clear = False):
		'''
			按顺序运行指令列表

			clear: 是否在运行完后清空指令列表
		'''
		self.start()
		for x in self.cm_lis:
			self.run_cm(*x)
		if clear:
			self.clear()

	def clear(self):
		'''
			清空指令列表
		'''
		self.cm_lis = []

if __name__ == "__main__":
	cmd = Cmd()
	cmd.add_cm("python && python\n")
	cmd.add_cm("print ('hello')\n")
	cmd.run()


from ..universe.beautiful_str import beautiful_str , merge_str
import os

def processer_quit(ui , cmd):
	ui.closed = True
def processer_addenv(ui , cmd):
	ui.env[cmd[1]] = cmd[2]
def processer_delenv(ui , cmd):
	ui.env.pop(cmd[1])

class CUI:
	def __init__(self , processers = [ 
			["q"   , processer_quit   , "quit"] , 
			["add" , processer_addenv , "add a variable"] , 
			["del" , processer_delenv , "delete a variable"] , 
		]):
		self.closed = False
		self.processers = {}
		for name , func , info in processers:
			self.add_processor(name , func , info) 

		self.env = {} # 用来记录中间变量

		# 记录最近一次的命令
		self.original_command = "" 	# 原始命令
		self.cmd = [] 				# 分词之后的命令
		self.cmd_var = {} 			# 解释之后的命令

		# 记录以往的所有输入和输出
		self.buffer = []

		# 记录预先处理的input
		self.pre_comms = []

	def buf(self , x):
		x = str(x)
		self.buffer.append(x)
		self.buffer = self.buffer[-1000:]


	def output(self , x = ""):
		'''输出一行'''
		self.buf(x)
		print (x)
	def input(self , x = ""):
		'''输入一行'''
		if len(self.pre_comms) > 0:
			inp = self.pre_comms.pop(0)
		else:
			inp = input(x)

		self.buf(x + inp)
		return inp

	def add_processor(self , name , func , info = ""):
		'''添加一个处理方法'''
		self.processers[name] = [ func , info ]

	def get_command(self):
		'''获取一行命令'''
		self.command = self.input(">> ")
		return self.command

	def warn(self , text = ""):
		'''输出警告'''
		self.output ("WARN: {text}".format(text = text))

	def interpret_cmd(self , cmd , names , types = None , update_env = False):
		'''根据给定的name和type，把cmd的list转成dict'''
		if len(cmd) < len(names):
			self.warn("Number of arguments is wrong")
			return
		assert types is None or len(types) >= len(names)

		var = {names[i] : types[i](cmd[i]) if types is not None else cmd[i] for i in range(len(names))}
		self.cmd_var = var
		if update_env:
			self.env.update(var)
		return var

	def get_env_str(self):
		return beautiful_str(
			["Var Name" , "Var Val"] , 
			[[name , val] for name , val in self.env.items()]
		).strip()

	def get_info_str(self):
		return beautiful_str(
			["Comm Name" , "Comm info"] , 
			[[name , info] for name , (_ , info) in self.processers.items()]
		).strip()

	def flush(self):
		# 刷新行为
		os.system("cls")
		env_str = self.get_env_str()
		info_str = self.get_info_str()
		op_str = merge_str(env_str , info_str , ensured_length = True)
		print ("\n".join(self.buffer)) #把最近的记录一起输出
		print (op_str)

	def run(self):
		while not self.closed:
			self.flush()

			cmd = self.get_command().strip()
			cmd = cmd.split(" ")

			if len(cmd) <= 0:
				continue

			self.cmd = cmd


			p = self.processers.get(cmd[0])
			if p is None:
				self.warn("No processer of {name}".format(name = cmd[0]))
			else:
				p[0](self , cmd) # p[0]: func

	def fake_input(self , comm):
		self.pre_comms.append(comm)


import math as M
from .strlen import max_len as real_len
def beautiful_str(name_list , value_list):
	'''输出一个很好看的表格。

	用法：
		beatiful_str(
			[title_1 , title_2 , ...] , 
			[
				[val_1_1 , val_1_2 , ... ] , 
				[val_2_1 , val_2_2 , ... ] , 
			]
		)

	示例：
		print (beautiful_str(
			["value_1" , "v2" , "value_3"] , 
			[
				[3.14156 , 2344 , 4.5678],
				[4,5,6],
				[123,"hahahaha" , 3456,],
			]
		))
	'''
	vn = len(name_list)

	len_list = list(name_list)
	for i in range(vn):
		len_list[i] = real_len(str(name_list[i]))
		for x in value_list:
			len_list[i] = max(len_list[i] , real_len(str(x[i])))
		len_list[i] += 2

	def get_imm(v):
		for i in range(vn):
			v[i] = str(v[i])

		the_str = "│"

		for i in range(vn):
			the_str += (" "*M.floor((len_list[i]-real_len(v[i]))/2))+v[i]+(" "*M.ceil((len_list[i]-real_len(v[i]))/2))
			the_str += "│"

		return the_str

	ceil , imm , floor = "┌" , "├" , "└"
	for i in range(vn):
		ceil  += ("─"*(len_list[i]//2))
		imm   += ("─"*(len_list[i]//2))
		floor += ("─"*(len_list[i]//2))
		if i != vn-1:
			ceil  += "┬"
			imm   += "┼"
			floor += "┴"
		else:
			ceil  += "┐"
			imm   += "┤"
			floor += "┘"

	title = get_imm(name_list)
	cont = ""
	for x in value_list:
		cont += get_imm(x) + "\n"

	return "\n" + ceil + "\n" + title + "\n" + imm + "\n" + cont + floor


def merge_str(s1 , s2 , inter = " " , ensured_length = False):
	'''
	给定两个带换行的字符串，把它们逐行合并
	TODO：自动判断字符串长度的方法有问题
	'''
	s1_line_len = real_len(s1) #s1每行的最长长度
	s1 , s2 = s1.split("\n") , s2.split("\n")
	s3 = ""
	for _i in range(max(len(s1),len(s2))):
		if not ensured_length:
			now_s = "" if _i >= len(s1) else s1[_i]
			now_s += " " * (s1_line_len - real_len(now_s))
		else:
			now_s = (" " * len(s1[0])) if _i >= len(s1) else s1[_i]

		if _i < len(s2): 
			now_s += inter + s2[_i]
		s3 += now_s + "\n"
	return s3

if __name__ == "__main__":
	print (beautiful_str.__doc__)
	print (beautiful_str(
		["value_1" , "v2" , "value_3"] , 
		[
			[3.14156 , 2344 , 4.5678],
			[4,5,6],
			[123,"hahahaha" , 3456,],
		]
	))




from .watch_time import my_clock
from ..universe.strlen import last_len
import time

class Logger:
	'''自动日志。

	参数：
	mode：一个函数，代表输出方向。"write"表示向文件写
	log_path：日志文件的位置。如果选了"write"的话必须填此项。如果打开了文件最后一定要调用close()
	append：往末尾添加哪些信息。clock表示当前运行秒数，time表示当前时间，也可以用其他函数


	方法：
	log：输出一个字符串
	close：关闭文件
	
	'''
	def __init__(self , mode = [print] , log_path = None , append = ["clock"] , line_length = 90):
		if log_path:
			self.log_fil = open(log_path , "w" , encoding = "utf-8")
		else:
			self.log_fil = None

		self.mode = mode

		if ("write" in mode) and (not log_path):
			raise Exception("Should have a log_path")

		self.append 	= append
		self.line_length = line_length

	def close(self):
		if self.log_fil:
			self.log_fil.close()

	def log(self , content = ""):

		content = self.pre_process(content)

		for x in self.mode:
			if x == "write":
				self.log_fil.write(content + "\n")
				self.log_fil.flush()
			else:
				x(str(content))

	def add_line(self , num = -1 , char = "-"):
		'''输出num个char , 如果num<0则默认line_length个
		'''
		if num < 0:
			num = self.line_length
		self.log(char * num)

	def pre_process(self , content):
		insert_space = self.line_length - last_len(content) #补全到line_length 
		content += " " * insert_space

		for x in self.append: #要往末尾添加哪些东西

			y = ""
			if x == "clock":
				y = "%.2fs" % (my_clock())
			elif x == "time":
				y = time.strftime("%Y-%m-%d %H:%M:%S" , time.localtime() )
			else:
				y = x()

			content += "| " + y + " "


		return content


import time
import sys
import pickle
from ..fakepath import new_fakefolder , fakepath_abs
from .filelock import FileLock
import os.path as P
import os
class LockerClient:

	NAME = "YLocker"
	VALFILE = "val.pkl"
	EMPTY_FLAG = "_EMPTY"

	def __init__(self):
		pass

	def getfolder(self , foldername):
		# 给定目录名，返回其绝对路径
		return new_fakefolder(P.join(self.NAME , foldername))

	def ensurefile(self , foldername):
		# 给定目录名，返回值文件的绝对路径
		filename = self.VALFILE
		path = P.join(self.getfolder(foldername) , filename)
		open(path , "ab").close()
		return path

	def encode(self , val):
		return pickle.dumps(val)
	def decode(self , content):
		if len(content) == 0:
			return None
		return pickle.loads(content)

	def get(self , key):
		path = self.ensurefile(key)
		with open(path , "rb") as fil:
			content = fil.read()
		return self.decode(content)

	def set(self , key , val):
		path = self.ensurefile(key)
		with FileLock(self.getfolder(key)): #锁定然后写入
			with open(path , "wb") as fil:
				fil.write(self.encode(val))
		return True
		
	def remove(self , key):
		path = self.ensurefile(key)
		with FileLock(self.getfolder(key)): #锁定然后写入
			os.remove(path)
		return True

	def plus(self , key , val = 1):
		'''同步加一，返回加后的值'''		
		path = self.ensurefile(key)
		with FileLock(self.getfolder(key)): #锁定然后加一
			with open(path , "rb") as fil:
				store_val = self.decode(fil.read())
			store_val = store_val + val
			with open(path , "wb") as fil:
				fil.write(self.encode(store_val))

		return store_val
		
	def set_if(self , key , expect_val , set_val):
		'''如果当前值 = expect_val，则设为set_val'''
		path = self.ensurefile(key)
		with FileLock(self.getfolder(key)):
			with open(path , "rb") as fil:
				store_val = self.decode(fil.read())
			if store_val == expect_val:
				with open(path , "wb") as fil:
					fil.write(self.encode(set_val))

		return True

	def _list_folders(self , folder):
		'''列出所目标子文件夹'''
		ret = []
		subfolers = os.listdir(folder)
		if self.VALFILE in subfolers: #存在值文件的是一个目标节点
			ret.append(folder)
		for p in subfolers:
			p = P.join(folder,p)
			if P.isdir(p):
				ret = ret + self._list_folders(p)
		return ret

	def ask_prefix(self , prefix , return_abs = False , only_suffix = False , not_none = False):
		'''查询所有key的前缀是给定prfix的key
		not_none：只返回非None的元素
		'''
		path = self.getfolder(prefix)
		ret = self._list_folders(path) #询问前缀下面的所有文件夹
		if not_none:
			ret = [x for x in ret if self.get(x) is not None]
		if return_abs:
			return ret
		if only_suffix:
			ret = [P.relpath(x , path) for x in ret] #求相对与前缀的相对路径
		else:
			the_path = self.getfolder("")
			ret = [P.relpath(x , the_path) for x in ret] #求伪路径
		return ret

	def clear(self):
		'''清除所有key'''
		all_dirs = self.ask_prefix("" , return_abs = True)
		for p in all_dirs:
			self.remove(p)

			if len( os.listdir(p) ) == 0: #为空才删除文件夹，否则就保留
				os.removedirs( p )
		self.clear_dir()
		return True

	def clear_dir(self , path = None):
		'''清除所有空文件夹'''
		if path is None:
			path = self.getfolder("")
		flag = True #自己是否能清除
		all_dirs = [P.join(path,p) for p in os.listdir(path) ]
		for p in all_dirs:
			if P.isdir(p):
				flag = flag and self.clear_dir(p)
			else:
				flag = False
		if flag: #现在应该已经没有文件了
			try:
				os.removedirs( path )
			except Exception:
				flag = False #不让删就算了
		return flag




import os
import os.path as P 

my_path = P.abspath( P.dirname(__file__) )
fakep_name = "files/"
fakep_path = P.join(my_path , fakep_name)

def ensure_fakep():
	if not P.exists(fakep_path):
		os.makedirs(fakep_path , exist_ok = True)

def new_fakefolder(path):
	'''在伪路径中创建目录，返回创建好的目录的绝对路径'''
	ensure_fakep()

	path = P.join(fakep_path , path)
	os.makedirs(path , exist_ok = True)

	return P.abspath(path)

def fakepath_abs(path):
	'''给定伪路径，返回绝对路径，不保证目录存在'''
	ensure_fakep()
	return P.abspath( P.join(fakep_path , path) )



import os
import os.path as P
import pickle
from .filename import hash

def try_load(key, path = ".", force_reload = False):
	if force_reload:
		return None
	filename = P.join(path , str(hash(key)))

	if P.exists(filename):
		with open(filename , "rb") as fil:
			data = pickle.load(fil)
		return data
	return None

def save_data(key , data , path = "."):
	filename = P.join(path , str(hash(key)))
	with open(filename , "wb") as fil:
		pickle.dump(data , fil)



'''求字符串的显示长度'''

def len_ignore_n(s):
	'''无视\\n的s的长度
	'''
	s = str(s).strip()
	s = s.replace("\n" , "")

	l = (len(bytes(s , encoding = "utf-8")) - len(s)) // 2 + len(s) #中文
	l += 7 * s.count("\t")											#\t

	return l

def max_len(s):
	'''每行最大长度
	'''
	s = str(s).strip()
	return max([len_ignore_n(x) for x in s.split("\n")])

def last_len(s):
	'''最后一行的长度
	'''
	s = str(s).strip()
	return len_ignore_n(s.split("\n")[-1])



from ..universe.extra_type import Proxy

class FileIntercepter(Proxy):
	'''向原文件写的时候，同时向另一个地方写'''
	def __init__(self , original , write_other):
		super().__init__(original = original , specials = ["write" , "_write_other"])
		self._write_other = write_other

	def write(self , s):
		self._original.write(s)
		self._write_other(s)



'''来自定义应用层协议吧！'''

def bytes2str(x):
	return str(x , encoding = "utf-8")

def str2bytes(x):
	return bytes(x , encoding = "utf-8")

def bytes2int(x):
	return int.from_bytes(x , byteorder = 'little' , signed = True)

def int2bytes(x , length = 4):
	return x.to_bytes(length , byteorder = 'little' , signed = True)

def int2ip(x):
	a = ((x & 0x000000FF) >> 0)
	b = ((x & 0x0000FF00) >> 8)
	c = ((x & 0x00FF0000) >> 16)
	d = ((x & 0xFF000000) >> 24)
	return "%d.%d.%d.%d" % (d,c,b,a)

def ip2int(x):
	d,c,b,a = [int(w) for w in x.strip().split(".")]
	return (d << 24) | (c << 16) | (b << 8) | (a << 0)

def bytes2ip(x):
	return int2ip(bytes2int(x))
	
def ip2bytes(x):
	return int2bytes(ip2int(x) , 4)

class DataReader:
	def __init__(self , data = None):
		self.data = data
		self.reset()

	def reset(self):
		self.now_pos = 0

	def read(self , length , data = None):
		'''从给定数据中读取特定长度的数据，如果length=-1，则会读取剩下所有数据
		'''
		if data is None:
			data = self.data

		if length < 0:
			return data[self.now_pos : ]

		self.now_pos += length
		return data[self.now_pos - length : self.now_pos]

	def __call__(self , *pargs , **kwargs):
		return self.read(*pargs , **kwargs)

class Protocol:
	def __init__(self , standard = None , head = None , body = None):
		'''
			协议标准：
			{
				"head": [
					[ 项目名 , 项目长度 , 编码方法 , 解码方法 ] ,
				] , 
				"body": [
					[ 项目名 , 项目长度 , 编码方法 , 解码方法 ] , 
				] , 
			}
			其中head的长度必须是固定的。body的项目长度可以是固定的，也可以是字符串，表示用长度为对应的head。
			body的最后一项也可以是-1，表示剩下所有数据

			例：
			standard = {
				"head": [
					["id"  , 4 , int2bytes , bytes2int] , 
					["len" , 4 , int2bytes , bytes2int] , 
				] , 
				"body": [
					["content" , "len" , str2bytes , bytes2str]
				]
			}

			也可以写作head = hhh , body = bbb。会被自动解释成standard = {"head" : hhh , "body": bbb}
		'''

		if standard is None:
			self.standard = {
				"head" : head , 
				"body" : body , 
			}
		else:
			self.standard = standard
		
		self.head = self.standard["head"]
		self.body = self.standard["body"]

		self.headsize = sum([x[1] for x in self.head])

	def encode_head(self , **kwargs):
		head = b''
		for name , length , encode_f , decode_f in self.head:
			encoded = encode_f(kwargs[name])

			if len(encoded) != length:
				raise "length incorrect"

			head += encoded
		return head

	def encode_body(self , **kwargs):
		body_lens = {}
		body = b''
		for name , length , encode_f , decode_f in self.body:
			encoded = encode_f(kwargs[name])

			if isinstance(length , int):
				if len(encoded) != length:
					raise "length incorrect"
			else:
				body_lens[length] = len(encoded)

			body += encoded

		return body , body_lens

	def encode(self , **kwargs):
		body , body_lens = self.encode_body(**kwargs)
		head = self.encode_head(**kwargs , **body_lens)
		return head + body

	def decode_head(self , head):

		decoded_head = {}

		data = DataReader(head)
		for name , length , encode_f , decode_f in self.head:
			encoded = data.read(length)
			decoded = decode_f(encoded)

			decoded_head[name] = decoded
		return decoded_head

	def decode_body(self , body , decoded_head):
		decoded_body = {}

		data = DataReader(body)
		for name , length , encode_f , decode_f in self.body:

			if isinstance(length , str):
				length = decoded_head[length]

			encoded = data.read(length)
			decoded = decode_f(encoded)

			decoded_body[name] = decoded
		return decoded_body


	def decode(self , encoded):
		head = encoded[:self.headsize]
		body = encoded[self.headsize:]

		decoded_head = self.decode_head(head)
		decoded_body = self.decode_body(body , decoded_head)

		decoded_body.update(decoded_head)

		return decoded_body



import random
import pickle

def random_name(lenmax = 8):
	g1 = "aeiou"
	g2 = "qrtpsdfghjkzxcvbnmw"
	g3 = "csdbh"
	g = [g1 , g2 , g3]
	l = random.randint(int(lenmax/2+0.5) , lenmax)
	last = random.randint(0,1)
	targ = ""
	for i in range(l):
		if last == 0: #在元音的情况下有0.2的概率保持不变
			if random.random() < 0.2:
				last = last ^ 1
		if last == 1: #在辅音的情况下有0.2的概率再来一个辅音
			if random.random() < 0.2:
				last = 2

		last = int(last != 0) ^ 1
		targ += random.sample(g[last] , 1)[0]

	return targ

def hash(data , mod = 114514114514):
	data = pickle.dumps(data)
	data = b"yyy" + data + b"ishandsome"
	hashed = int.from_bytes(data , byteorder = "little")
	if mod > 0:
		hashed = hashed % mod
	return hashed

def strhash(data):
	data = hash(data)
	

if __name__ == "__main__":
	print (random_name())
	print (hash(None))
	print (hash(None))
	print (hash({0:[print , "oooo"] , "a" : {} , "c": {None : {233}}}))
	p = "oooo"
	d = "a"
	print (hash({0:[print ,      p] ,   d : {} , "c": {None : {233}}}))
	print (hash({d : {} , 0:[print ,      p]   , "c": {None : {233}}}))




'''
	一个用来调试带有递归代码的工具
	可以显示递归逻辑、查找节点和查看节点的一些值
	
	其中Node表示节点，由Debuger创建
	RecursionDebuger是主工具，应为每一个需要调试的函数创建一个该类对象
'''

import re

class Node:
	'''
			father是一个Node，表示节点的父节点，为None则说明此节点为根节点
			name是一个元组，表示节点的名字，应是一个元组
			dic是一个字典，表示这个节点的可查询值列表
			son是一个字典，表示所有节点的名字和对应的节点
	'''
	def __init__(self , father = None , name = None , dic = {} , name_in_fat = None , name_in_all = None):
		'''
			father表示节点的父节点，为None则说明此节点为根节点
			name表示节点的名字，应是一个元组
			dic表示这个节点的可查询值列表
		'''
		self.father = father
		self.sons = {}
		self.name = name
		self.name_in_fat = name_in_fat
		self.name_in_all = name_in_all
		self.dic = dic

	def is_root(self):
		return self.father == None

	def ask_sons(self,p):
		'''
			查询所有名字为p的儿子列表
			返回值第一个元素是儿子节点对象，第二个元素是这个儿子在父亲下的名字
		'''
		ret = []
		if(self.sons.get(p)):
			ret.append( self.sons[p] )

		now_num = 0
		while True:
			the_tuple = (p , now_num)
			if(self.sons.get(the_tuple)):
				ret.append( self.sons[the_tuple] )
			else:
				break
			now_num += 1
		return ret

	def add_son(self,node):
		name = node.name
		if(self.sons.get(name) != None):
			name = (name,0)
		while(self.sons.get(name) != None):
			name = (name[0],name[1] + 1)
		self.sons[name] = node
		return name

class RecursionDebuger_CUI:
	def __init__(self,parent):
		self.parent = parent
		self.now_node = parent.root

	def get_node_name_info(self,node):

		fat_nam = "root"
		if(node.father != None):
			fat_nam = node.father.name_in_all
		return "{0} ，父亲为：{1} ，父中名为：{2}，全局名为：{3}".format(
				node.name , 
				fat_nam , 
				node.name_in_fat , 
				node.name_in_all 
			)

	def print_node(self,node,start_info = "" , end_info = ""):

		print ()
		print (start_info)
		print ("------------")

		print("当前节点 " + self.get_node_name_info(node) + "，信息列表：")

		for ite in node.dic:
			print ("\t" + str(ite) + " : " + str(node.dic[ite]))

		print ("------------")
		print (end_info)
		print ()

	def go_into_son(self,node,start_info = "" , end_info = ""):

		print ()
		print (start_info)
		print ("------------")

		name = input("输入儿子的名：")
		name = re.split("," , name)
		for i in range(len(name)):
			name[i] = int(re.sub("\D" , "" , name[i]))
		name = tuple(name)

		ret = node.ask_sons(name)
		while(len(ret) == 0):
			print ("儿子不存在，重新输入...")
			name = input("输入儿子的名：")
			if(len(name) == 0):
				print ("退出输入...")
				return 
			name = re.split("," , name)
			for i in range(len(name)):
				name[i] = int(re.sub("\D" , "" , name[i]))
			name = tuple(name)
			ret = node.ask_sons(name)

		node_num = -1
		if(len(ret) == 1):
			node_num = 0
		elif(len(ret) > 1):
			print("有若干匹配的节点，输入编号来指定具体某一个：")

			_cnt = 0
			for the_son in ret:
				print("{0} : {1}".format(_cnt , self.get_node_name_info(the_son)))
				_cnt += 1

			node_num = -1
			while(node_num < 0 or node_num >= len(ret)):
				node_num = int(input("输入编号："))

		self.now_node = ret[node_num]

		print("已进入。节点名:{0}".format(self.now_node.name))

		print ("------------")
		print (end_info)
		print ()

	def look_sons(self,node,start_info = "" , end_info = ""):

		print ()
		print (start_info)
		print ("------------")

		print ("当前节点 {0} 儿子列表：".format(node.name))

		for nam in node.sons:
			print ("\t" + str(nam))

		print ("------------")
		print (end_info)
		print ()

	def goto_father(self,node,start_info = "" , end_info = ""):

		print ()
		print (start_info)
		print ("------------")

		self.now_node = self.now_node.father

		print("完成。")
		print("当前节点名:{0}".format(self.now_node.name))

		print ("------------")
		print (end_info)
		print ()


	def exec_(self):

		self.now_node = self.parent.root

		commad_list = {
			"当前节点" 		: 0,
			"0" 			: 0,
			"info" 			: 0,

			"进入儿子" 		: 1,
			"1" 			: 1,
			"down" 			: 1,

			"查看儿子" 		: 2,
			"2" 			: 2,
			"son" 			: 2,

			"查找名字" 		: 3,
			"3" 			: 3,
			"search" 		: 3,

			"进入父亲" 		: 4,
			"4" 			: 4,
			"up" 			: 4,

		}

		while(True):

			print()
			cm = input("输入指令：")
			while(commad_list.get(cm) == None):
				print ("指令 \"{0}\" 不在指令列表中。".format(repr(cm)))
				print (cm)
				cm = input("输入指令：")

			cm = commad_list[cm]

			if(cm == 0):
				self.print_node(self.now_node)
			elif(cm == 1):
				self.go_into_son(self.now_node)
			elif(cm == 2):
				self.look_sons(self.now_node)
			elif(cm == 3):
				self.go_into_son(self.parent.all)
			elif(cm == 4):
				self.goto_father(self.now_node)

class RecursionDebuger:

	def __init__(self):
		'''
			·在递归过程的开始调用set_point()以创建节点，并给定一个元组作为节点名
				（可以重名，若如此会被加入额外的编号作为命名（比如(a,b)会被改为((a,b),0)））

			·在递归结束时调用end_point()以结束创建节点，并给定一个dict作为可查找变量列表

			在递归完整结束后，可以通过这个工具来查看所记录的信息
		'''
		self.stack = []
		self.root = None
		self.now_info = {}
		self.all = Node()


	@property
	def cui(self):
		'''
			创建本调试器的命令行界面
			需要在信息获取完全后创建
		'''
		return RecursionDebuger_CUI(self)

	def set_point(self,name):
		'''
			创建节点
			name表示将要创建的节点的名字
		'''
		fat = None
		if(len(self.stack) > 0):
			fat = self.stack[-1]

		new_node = Node(father = fat , name = name , dic = {})

		if(fat):
			the_name = fat.add_son(new_node)
			new_node.name_in_fat = the_name
		else:
			self.root = new_node

		the_name = self.all.add_son(new_node)
		new_node.name_in_all = the_name

		self.stack.append(new_node)

	def add_info(self,info_name,info_val):
		#TODO : BUG!!!!
		self.now_info[info_name] = info_val

	def end_point(self,dic = {}):

		if(len(self.stack) <= 0):
			raise Exception("RecursionDebuger.end_pont() : bad call")

		the_las = self.stack.pop()

		self.now_info.update(dic)
		the_las.dic = self.now_info
		self.now_info = {}





import os
import django
from django.conf import settings
from django.core.management import execute_from_command_line
from django.urls import path
from pathlib import Path
from .urlconf import add_response

def start_server(ip = "0.0.0.0", port = "30726", responsers = {} , encode = "pickle" , cross_domain = False):

	configs = {
		"encode" : encode , 
		"cross_domain": cross_domain , 
	}

	for url , func in responsers.items():
		add_response(url , func , configs)

	settings.configure(
		DEBUG = True,	
		ROOT_URLCONF = "YTools.network.server.urlconf", 
		SECRET_KEY = "THEWORLD", # unsafe!!
	)
	django.setup()

	ipaddr = "{ip}:{port}".format(ip = ip , port = port)
	execute_from_command_line(["" , "runserver" , ipaddr])


'''
	一个GLSL使用库
	使用ShaderProgram类来创建着色器程序
	使用Shader类来创建着色器

	沒有什麼錯誤判斷，不正確使用的話會爆炸吧大概

	TODO：最近一次测试发现不会用了
'''

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL.GL import shaders
import types

class Shader:
	'''
		用来创建和保存一个着色器对象的类
	'''
	def __init__(self,shader_type = None , file_name = None , string = None , auto_build = True):
		'''
			从一个文件名加载，或是从一个现成的字符串。
			如果都提供了，则优先从文件名加载
			
			shader_type	：着色器类型，一个OpenGL常量
			file_name	：文件名
			string 		：字符串
			auto_build	：是否自动编译
		'''
		if file_name != None:
			self.set_data_file(file_name)
		else:
			self.set_data_string(string)

		#着色器类型
		self.shader_type = shader_type

		if auto_build:
			self.build()

	def load_from_file(self,file_name,encoding = "utf-8"):
		'''
			从文件读取一个字符串
		'''
		return open(file_name,encoding=encoding).read()

	def set_data_file(self,file_name):
		'''
			从文件加载着色器内容
		'''

		#着色器内容（源代码）
		self.data_str = self.load_from_file(file_name)

	def set_data_string(self,string):
		'''
			从字符串加载着色器内容
		'''
		self.data_str = string

	def build(self):
		'''
			创建着色器对象并编译着色器源代码
		'''

		if not self.data_str:
			raise Exception("Shader.compile() : no data")
		self.shader_object = glCreateShader(self.shader_type)
		glShaderSource(self.shader_object , self.data_str)
		glCompileShader(self.shader_object)



class ShaderProgram:
	
	def __init__(self , files = [] , strings = [] , end_create = True , use = True):
		'''
			一个着色器程序类
			可以预先指定一系列着色器内容，来添加着色器

			files		：二元组列表，每个元素为一个着色器信息，将从这些信息来自动创建着色器
							每个二元组，第一个元素为着色器类型，第二个元素为文件名
			strings		：同上，不过每个二元组的第二个元素为源代码字符串
			end_create	：是否链接这个程序
			use 		：是否激活这个程序
		'''
		self.program =  glCreateProgram()

		if files:
			for shader_type , file_name in files:
				self.add_shader_from_file(shader_type , file_name)

		if strings:
			for shader_type , string in strings:
				self.add_shader_from_string(shader_type , string)

		if end_create:
			self.end_create()

		if use:
			self.enable()

	def end_create(self):
		'''
			链接
		'''
		glLinkProgram(self.program)

	def enable(self):
		'''
			激活自己这个着色器程序
		'''
		glUseProgram(self.program)

	def disable(self):
		'''
			关闭这个程序。
			由于实际上只有一个着色器程序能处于激活状态，所以这个函数会关掉这个全局的活跃着色器，不管他是不是自己
		'''
		glUseProgram(0)

	def add_shader(self,shader):
		'''
			绑定一个着色器对象
		'''
		glAttachShader(self.program , shader.shader_object)

	def add_shader_from_string(self,shader_type,string):
		'''
			从字符串自动创建一个着色器对象，并自动绑定
		'''
		self.add_shader ( Shader(shader_type , string = string) )

	def add_shader_from_file(self,shader_type,file_name):
		'''
			从文件自动创建一个着色器对象，并自动绑定
		'''
		self.add_shader ( Shader(shader_type , file_name = file_name) )

	def update_projection_matrix(self,name):
		'''
			更新顶点着色器中的投影矩阵

			name：全局变量名
		'''
		self.set_uniform_matrix("mat_pro", glGetFloatv(GL_PROJECTION_MATRIX) , float)

	def update_modelview_matrix(self,name):
		'''
			更新建顶点着色器中的模观察矩阵

			name：全局变量名
		'''
		self.set_uniform_matrix("mat_modview", glGetFloatv(GL_MODELVIEW_MATRIX) , float)


	def set_uniform(self,uni_name,val,type_name = float):
		'''
			设置着色器中的全局变量

			uni_name	：全局变量变量名
			type_name 	：全局变量元素类型（float或者int）
			val			：一个包含了全局变量的列表，应和全局变量（如果是vec）的话的长度一致
		'''
		the_uni = glGetUniformLocation(self.program , uni_name);
		the_source = "glUniform{0}{1}(the_uni , *val)".format(len(val),type_name.__name__[0])
		exec(the_source)

	def set_uniform_matrix(self,uni_name,val,type_name = float):
		'''
			设置着色器中的全局变量

			uni_name	：全局变量变量名
			type_name 	：全局变量元素类型（float或者int）
			val			：一个包含了全局变量的列表，应和全局变量（如果是vec）的话的长度一致
		'''
		the_uni = glGetUniformLocation(self.program , uni_name)
		if len(val) == len(val[0]):
			the_source = "glUniformMatrix{0}{1}v(the_uni , 1 ,False, val)".format( len(val) , type_name.__name__[0])
		else:
			the_source = "glUniformMatrix{0}x{1}{2}v(the_uni , 1 ,False, val)".format( len(val) , len(val[0]) , type_name.__name__[0])
		exec(the_source)






import random

def set_random_seed(seed , modules = ["random" , "torch" , "numpy"]):
	'''设置随机种子
		
	参数：
	seed：随机种子的值

	modules：要为哪些模块设置随机种子
		目前可选：random、torch、numpy
		注意torch中cuda也会被设置。
		只要gpu没问题，可以放心的使用。
	'''
	if "random" in modules:
		random.seed(seed)
	if "numpy" in modules:
		import numpy as np
		np.random.seed(seed)
	if "torch" in modules:
		import torch as tc
		tc.manual_seed(seed)
		tc.cuda.manual_seed_all(seed)
		tc.backends.cudnn.deterministic = True
		tc.backends.cudnn.benchmark = False



from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE', encoding='utf-8') as f:
    license = f.read()

with open('requirements.txt', encoding='utf-8') as f:
    reqs = f.read()

pkgs = [p for p in find_packages() if p.startswith('YTools')]
print(pkgs)

setup(
    name='YYYTools',
    version='0.1.6',
    url='http://github.com/FFTYYY/YTools',
    description='',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='MIT',
    author = 'Yongyi Yang',
	author_email = 'yongyi@umich.edu',
    python_requires = '>=3.6',
    packages = pkgs,
    install_requires = reqs.strip().split('\n'),

    include_package_data = True , 
)


import pdb
from .beautiful_str import beautiful_str

class Struct(object):		
	'''一个可以自由添加元素的类。
	'''

	def __init__(self , _default = None , **kwargs):
		self.__dict__.update(kwargs)
		self._default = _default
		self._getter = {}
		self._setter = {}

	def _set_property(self , name , getter , setter):
		self._getter[name] = getter
		self._setter[name] = setter

	def __getattr__(self , name):

		if name.startswith("_"):
			return object.__getattribute__(self , name)

		if hasattr(self , "_getter") and self._getter.get(name):
			return self._getter[name](self)

		if name not in self.__dict__:
			self.__dict__[name] = self._default

		return self.__dict__.get(name)

	def __setattr__(self , name , val):

		if hasattr(self , "_setter") and self._setter.get(name):
			self._setter[name](self , val)
			return 

		self.__dict__[name] = val

	def __str__(self):
		return beautiful_str(["name" , "value"] , [[x , self.__dict__[x]] for x in self.__dict__ if not x.startswith("_")])

SPECIAL = "_SPECIAL"

class Proxy:
	'''一个可以重写其他元素的类
	'''

	def __init__(self , original , specials):
		'''
		original：代理对象
		specials：代理属性


		访问一个Proxy对象时，如果属性在specials中，则会返回一个新版本，对于其他属性则都返回original的版本
		修改specials属性不影响original，其他属性则影响
		因此，可以重写specials属性，其他属性直接继承original
		'''
		self.__dict__["_special"]  = ["_original"] + specials
		self.__dict__["_original"] = original

	def __getattr__(self , name):
		if name == "_special" or name == "__dict__": #直接查询本对象
			return object.__getattribute__(self , name)

		if name in self._special: #直接查询本对象
			if name in self.__dict__:
				return self.__dict__[name]

		return getattr(self._original , name)

	def __setattr__(self , name , value):

		if name in self._special or name == "_special": #直接从dict设置本对象
			self.__dict__[name] = value
			return

		return setattr(self._original , name , value)



import os

class TempCWD:
	'''临时CWD。

	用法：
	with TempCWD(path):
		do_someting

	然后这个do_something会被视为在path下运行，不会影响原始环境的cwd
	'''

	def __init__(self , temp_path):
		self.temp_path = temp_path

	def __enter__(self):
		self.old_cwd = os.getcwd()
		os.chdir(self.temp_path)
		return self

	def __exit__(self , *args , **kwargs):
		os.chdir(self.old_cwd)





'''这个模块用于管理程序的退出时行为。

全局变量:
	on_quit_list: 退出时要调用的方法列表，在退出时会依次调用（当然，不能有参数）。

方法:
	add_quit_process: 向on_quit_list中添加一个方法。
'''
import atexit

on_quit_list = []

def add_quit_methods(f):
	'''向on_quit_list中添加一个方法。
	'''
	on_quit_list.append(f)

def on_quit():
	for x in on_quit_list:
		x()

atexit.register(on_quit)